package net.sourceforge.ppf;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import net.sourceforge.ppf.util.Filter;
import net.sourceforge.ppf.util.PPFHelp;
import net.sourceforge.ppf.util.PPFResourceBundle;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Text;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.jdom.xpath.XPath;
import org.jibble.pircbot.DccChat;
import org.jibble.pircbot.DccFileTransfer;
import org.jibble.pircbot.User;

/**
 * Abstract base class for all PPF plugins.
 * @author DeadEd
 */
public abstract class PPFPlugin {
    
    private static final org.apache.commons.logging.Log LOG = org.apache.commons.logging.LogFactory
            .getLog(PPFPlugin.class);
    /**
     * Version number of the plugin
     */
    protected String PLUGIN_VERSION = "0.0"; 
    
    /**
     * Plugin ID
     */
    protected String PLUGIN_ID = "ABSTRACT";
    
    /**
     * Default timeout for a user to accept a dcc chat request (in milliseconds)
     */
    private int DEFAULT_DCC_CHAT_TIMEOUT = 120000;

    /**
     * Output will be sent to the channel
     */
	protected static final int OUTPUT_CHANNEL = 0;
	/**
	 * Output will be sent to a private message
	 */
	protected static final int OUTPUT_PM = 1;
	/**
	 * Output will be sent as a notice
	 */
	protected static final int OUTPUT_NOTICE = 2;

    /**
     * Instance of the PPF bot - used to access any PPF and PircBot methods 
     */
    private PPF bot = null;
    
    /**
     * Help texts for the plugin.  Keys are converted to lowercase for case insensitive lookup.
     */
    private HashMap help = new HashMap();
    
    /**
     * A resource bundle that contains all of the available language resource bundles for this plugin 
     */
    private PPFResourceBundle ppfResBundle;

    /**
     * Easy way to keep hold of a list of String items.  Methods for adding, removing and
     * viewing the items on this list are available to the plugin.  These methods also update the
     * XML document where the list is stored.  All entries and operations on the entries are done
     * in lowercase for case insensitive lookup.
     */    
    private HashMap itemList = new HashMap();

    /**
     * JDOM doc from the XML
     */
    private Document doc;
    
    /**
     * The directory where this plugin is running from
     */
    private String urlStr;
    
    /**
     * The filename of the XML doc
     */
    private String xmlFilename; 
    
    
    /**
	 * Constructor
	 */
	public PPFPlugin() {}
    	
    /**
     * Called by the core when a plugin is loaded.  
     * 
     * This method is called every time a plugin is loaded, which means at the initial loading 
     * and also when a plugin is reloaded (reloading means unload and then load).
     * 
	 * @param ppf the bot instance used to create this plugin
	 */
	protected void init(PPF ppf) {
        bot = ppf;
    }

	/**
	 * Returns the instance of PPF that this plugin has been initialised with.
	 * 
	 * @return PPF the PPF instance
	 */
	public PPF getBot() {
		return bot;
	}

    /**
     * Unload hook to place any clean-up code in - this is called by PPF when unloading the plugin.
     */
    protected void unload() {
    }

	/**
	 * Figure out where the output will go.
	 * 
	 * Where some commands output to is configurable. This takes the configured text
	 * and returns one of the plugin OUTPUT constants. 
	 * 
	 * @param output string representation of where the output will go
	 * @return PPF output from the string representation, defaults to OUTPUT_CHANNEL if a 
	 * value that is not recognised if given
	 */
	protected int outputFromString(String output) {		
		String out = output.toUpperCase().trim();

		if(out.equals("CHANNEL")) {
			return OUTPUT_CHANNEL;
		} else if(out.equals("PM")){
			return OUTPUT_PM;
		} else if(out.equals("NOTICE")) {
			return OUTPUT_NOTICE;
		}
		
		return OUTPUT_CHANNEL;
	}

	/**
	 * Sends the output to the place that was configured.
	 * 
	 * Using the OUTPUT constant, decides how to send the message from the bot.
	 * Messages may go to channel, private message, or notice, depending on the 
	 * OUTPUT constant used. Defaults to sending to the channel.
	 * 
	 * @param output the output method, one of the plugin OUTPUT constants
	 * @param channel the name of the channel to send to
	 * @param sender the name of the sender to send to
	 * @param message the message that will be sent
	 */
	public void sendOutput(int output, String channel, String sender, String message) {
		switch(output) {
			case OUTPUT_CHANNEL:
				bot.sendMessage(channel, message);
				break;
			case OUTPUT_PM:
				bot.sendMessage(sender, message);
				break;
			case OUTPUT_NOTICE:
				bot.sendNotice(sender, message);
				break;
			default:
				bot.sendMessage(channel, message);
		}
	}
	
    /////////////////////////////////////////////////////////////////
	//
	// Version
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Gets the version of the plugin.
	 * 
	 * This is used by the core when displaying version information about the plugin.
	 * 
	 * @return the plugin version
	 */
    public String getVersion() {
        return PLUGIN_VERSION;
    }

    /**
     * Set the version of the plugin.
     * 
     * This should be done by the plugin in the Constructor or init() method. 
     * 
     * @param thisVersion version of the plugin
     */
    protected void setVersion(String thisVersion) {
        PLUGIN_VERSION = thisVersion;
    }

	/////////////////////////////////////////////////////////////////
	//
	// Resource bundle
	//
	/////////////////////////////////////////////////////////////////

    /**
     * A resource bundle that contains all of the available language resource bundles for this plugin.
     * 
     * @return PPFResourceBundle bundle for this plugin
     */
    protected PPFResourceBundle getPPFResourceBundle() {
        return ppfResBundle;
    }
       
    /**
     * Read a set of properties files in for use by this plugin.
     * 
     * @param bundleName the base name of the properties to read in
     */
    protected void readProperties(String bundleName) {
        ppfResBundle = new PPFResourceBundle(bundleName);
        Filter filter = new Filter("properties");

        URL urli = getClass().getProtectionDomain().getCodeSource().getLocation();
        String urlStr = urli.getFile();
        if(!urlStr.endsWith("/")) {
            urlStr = urlStr.substring(0, urlStr.lastIndexOf("/")+1);
        }

        File dir = new File(urlStr);

        // get a list of all the properties files
        String[] strs = dir.list(filter);
        for(int i=0; i < strs.length; i++) {
            int start = strs[i].lastIndexOf("_") + 1;
            if(start != -1) {
                int end = strs[i].indexOf(".", start);
                if(end != -1) {
                    // strip the language code from the filename 
                    String langStr = strs[i].substring(start, end);
                    // add the bundle to the available bundles for this plugin 
                    ppfResBundle.addLang(langStr);
                }
            }
        }
    }

	/////////////////////////////////////////////////////////////////
	//
	// XML Document
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * The JDom Document that is created when reading in the XML config. 
	 * 
	 * @return Document the JDom Document representing the XML config.
	 */
    protected Document getDoc() {
	    return doc;
	}
	    
	/**
	 * The location of where this plugin is running from in URL from.
	 * 
	 * @return String url formatted location of where the plugin is running from.
	 */
	protected String getUrlStr() {
	    return urlStr;
	}
	    
	/**
	 * The filename of the XML config that is used by this plugin.
	 * 
	 * @return String the XML config filename
	 */
	protected String getXMLFilename() {
	    return xmlFilename;
	}

    /**
     * Sets the location of where this plugin is running from into the <code>urlStr</code> variable
     */
    private void setRuntimeRoot() {
        URL urli = getClass().getProtectionDomain().getCodeSource().getLocation();
        urlStr = urli.toString();
        if(!urlStr.endsWith("/")) {
            urlStr = urlStr.substring(0, urlStr.lastIndexOf("/")+1);
        }
    }

	/**
     * Loads an XML document.
     * 
     * This is generally for loading the plugin config files.
     * 
     * @param filename the XML document to load
     * @return Document a JDOM representation of the XML file read in
     */
    protected Document loadDoc(String filename) {
        setRuntimeRoot();
        
        xmlFilename = filename;
        
        SAXBuilder builder = new SAXBuilder();

        URL newURL;
        try {
            newURL = new URL(urlStr + xmlFilename);
            doc = builder.build(newURL);
        } catch (MalformedURLException e) {
            LOG.error("Error loading plugin configuration", e);
        } catch (JDOMException e1) {
            LOG.error("Error parsing plugin configuration", e1);
        } catch (FileNotFoundException e) {
        	LOG.error("Plugin configuration file not found: "+ filename);
            return null;
        } catch (IOException e1) {
            LOG.error("Error loading plugin configuration", e1);
        }

        return doc;
    }

	/**
	 * Get text from an XML element.
	 * 
	 * @param path the path to the element containing the text
	 * @return the content of the element as a String, null if not found
	 */
	protected String getTextFromDoc(String path) {
		String retVal = null;
		try {
			XPath xpQuery = XPath.newInstance(path + "/text()");
			retVal = ((Text) xpQuery.selectSingleNode(doc)).getTextTrim();
		} catch (Exception e) {
			LOG.error("Error in PPFPlugin.getTextFromDoc: problem using path: "+ path +"/text()", e);
			retVal = null;
		}
		
		return retVal;
	}

	/**
	 * Get a list of elements from the Document.
	 * 
	 * @param path the path to the element containing the list
	 * @return a List of the elements under the path, null if not found
	 */
	protected List getListFromDoc(String path) {
		List retVal = null;
		try {
			XPath xpList = XPath.newInstance(path);
			retVal = xpList.selectNodes(doc);
		} catch (Exception e) {
			LOG.error("Error in PPFPlugin.getListFromDoc: problem using path: "+ path, e);
			retVal = null;
		}
		return retVal;
	}
	
	/**
	 * Get the boolean value from a text representation.
	 * 
	 * Valid values that will return true are: yes, true, on.
	 * All other values will return false.
	 * 
     * @param path the path to the element containing the text
     * @return boolean value represented by the text
     */
    protected boolean getBooleanFromDoc(String path) {
		String val = getTextFromDoc(path);
		if (val.equalsIgnoreCase("yes")
			|| val.equalsIgnoreCase("true")
			|| val.equalsIgnoreCase("on")) {

			return true;
		}
		
		return false;
	}
	
	/**
	 * Get an attribute from an element in the document.
	 * 
     * @param path the path to the element containing the attribute
     * @param attribute the attribute to get the text from
     * @return the content of theattribute, null if not found
     */
    protected String getAttributeFromDoc(String path, String attribute) {
		String retVal = null;
        try {
			XPath xpQuery = XPath.newInstance(path);
			Element query = (Element) xpQuery.selectSingleNode(doc);
			retVal = query.getAttributeValue(attribute);
        } catch (Exception e) {
			LOG.error("Error in PPFPlugin.getAttributeFromDoc: problem using path: "+ path
				+" and attribute: "+ attribute, e);
			retVal = null;
        }
        
        return retVal;
	}

	/////////////////////////////////////////////////////////////////
	//
	// Help
	//
	/////////////////////////////////////////////////////////////////

    /**
     * Adds online help information for a command in the plugin.
     * 
	 * @param command the command to associate the help with
	 * @param ppfHelp the help text for the command
	 */
	protected void setHelp(String command, PPFHelp ppfHelp) {
        if(command == null) {
            return;         
        }
        if(ppfHelp == null) {
            return;
        }
        if(help.containsKey(command.toLowerCase())) {
            LOG.error("Help command already given: "+ command);
        } else {
            help.put(command.toLowerCase(), ppfHelp);
        }
    }
    
    /**
     * Gets all of the help for this plugin.
     * 
     * @return the help available for this plugin, the key is the command
     */
    protected HashMap getHelp() {
        return help;
    }

	/////////////////////////////////////////////////////////////////
	//
	// Item List
	//
	/////////////////////////////////////////////////////////////////
	
	/** 
	 * Easy way to keep hold of a list of String items.  
	 * 
	 * @return HashMap helper list that can be operated on easily with the XML config
	 */
	protected HashMap getItemList() {
		return itemList;        
	}
    
    /**
     * Select a single item from the list by using the key.
     * 
     * @param key to search for
     * @return the value, null if not in list
     */
    protected String getFromItemList(String key) {
    	if(isInItemList(key)) {
    		return (String) itemList.get(key.toLowerCase());
    	}
    	
    	return null;
    }

	/**
	 * Check if an item is in the list, by key.
	 * 
     * @param key to check for
     * @return true if the key is in the list
     */
    protected boolean isInItemList(String key) {
		if(itemList.containsKey(key.toLowerCase())) {
			return true;
		}
		
		return false;
	}

	/**
	 * Adds a new item to the list in memory.
	 * 
     * @param key to use for adding the item
     * @param value item to add to the list
     * @return true if the item was added, false if it already existed on the list
     */
    protected boolean addToItemList(String key, String value) {
		// only add if not already in the list
		if(!isInItemList(key)) {
			itemList.put(key.toLowerCase(), value);
			return true;
		}
		
		return false;
	}

    /**
     * Adds a new item to the persitent list.
     * 
     * @param listElem name of the list element
     * @param newElement new element to add
     * @param key to use for adding the item
     * @param value item to add to the list
     * @return true if the item was added, false if any problem occured while adding the item
     */
    protected boolean addToItemList(String listElem, Element newElement, String key, String value) {
        boolean retVal = true;

        FileOutputStream fos = null;
        try {
            Element el = doc.getRootElement().getChild(listElem);
            if(el == null) {
			    el = new Element(listElem);
			    doc.getRootElement().addContent(el);
            }
                        
            addToItemList(key, value);
    
            el.addContent(newElement);
            Format format = Format.getPrettyFormat();
            format.setTextMode(Format.TextMode.NORMALIZE);
            format.setLineSeparator(System.getProperty("line.separator"));

            fos = new FileOutputStream(new URL(urlStr + xmlFilename).getFile());
            XMLOutputter output = new XMLOutputter(format);
            output.output(doc, fos);

        } catch(FileNotFoundException ex) {
            retVal = false;
            LOG.error("Error opening plugin configuration file", ex);
        } catch(IOException ex) {
            retVal = false;
            LOG.error("Error writing to plugin configuration file", ex);
        } catch(Exception ex) {
            retVal = false;
            LOG.error("Error updating plugin configuration", ex);
        } finally {
            try {
                if(fos != null) {
                    fos.close();
                }
            } catch(IOException ex) {
                retVal = false;
                LOG.error("Error closing plugin configuration file", ex);
            }
        }
        
        return retVal;
    }

    /**
     * Remove an item from the persitent list.
     * 
     * @param childElem name of the list element
     * @param xPath path to the element
     * @param key key of the item to be removed
     * @return true if the item was removed, false if any problem occured while removing the item
     */
    protected boolean removeFromItemList(String childElem, String xPath, String key) {
        boolean retVal = true;
        if(isInItemList(key)) {
            FileOutputStream fos = null;
            try {
                itemList.remove(key.toLowerCase());
                        
                XPath xpRemove = XPath.newInstance(xPath);
                Element remove = (Element) xpRemove.selectSingleNode(doc);
                
                if (remove == null) {
                    LOG.error("Element to remove not found: xpath=" + xPath);
                    //TODO retVal = false; ?
                } else {
                    Element el = doc.getRootElement().getChild(childElem);
                    if (!el.removeContent(remove)) {
                        LOG.error("Element was not a child of " + childElem);
                        //TODO retVal = false; ?
                    } else {
                        Format format = Format.getPrettyFormat();
                        format.setTextMode(Format.TextMode.NORMALIZE);
                        format.setLineSeparator(System.getProperty("line.separator"));
    
                        fos = new FileOutputStream(new URL(urlStr + xmlFilename).getFile());
                        XMLOutputter output = new XMLOutputter(format);
                        output.output(doc, fos);
                    }
                }
                    
            } catch (JDOMException e1) {
                retVal = false;
                LOG.error("Error parsing plugin configuration", e1);
            } catch (IOException e1) {
                retVal = false;
                LOG.error("Error writing to plugin configuration file", e1);
            } finally {
                try {
                    if(fos != null) {
                        fos.close();
                    }
                } catch(IOException ex) {
                    retVal = false;
                    LOG.error("Error closing plugin configuration file", ex);
                }
            }    
        }
        
        return retVal;
    }

    /**
     * View all items in the list.
     * 
     * This uses DCC chat due to the possible volume of the list.
     * The default DCC chat timeout is used.
     * 
     * @param sender the person to create a DCC chat session with
     */
    protected void viewItemList(String sender) {
        viewItemList(sender, DEFAULT_DCC_CHAT_TIMEOUT);
    }
    
    /**
     * View all items in the list.
     * 
     * This uses DCC chat due to the possible volume of the list.
     * 
     * @param sender the person to create a DCC chat session with
     * @param timeout how long to wait for an answer before giving up
     */
    protected void viewItemList(String sender, int timeout) {
        //FIXME: from PircBot's javadoc: It is strongly recommended that you call this method within a new Thread, as it may take a long time to return. 
        DccChat chatSession = bot.dccSendChatRequest(sender, timeout);
        if (chatSession != null) {
            try {
                Iterator keys = itemList.keySet().iterator();
                while (keys.hasNext()) {
                    String key = (String) keys.next();
                    chatSession.sendLine(key + " - " + itemList.get(key));
                }

            } catch (IOException ex) {
                LOG.error("Error during DCC chat with " + sender, ex);
            } finally {
                try {
                    chatSession.close();
                } catch (IOException e) {
                    LOG.warn("Error closing DCC chat with: " + sender, e);
                }
            }
        }
    }

    /////////////////////////////////////////////////////////////////
	//
	// Auth
	//
	/////////////////////////////////////////////////////////////////

	/**
	 * Figure out what the required auth level is.
	 * 
	 * With some commands, you can configure who is allowed to run them.  This takes the 
	 * configured text and returns an auth level that PPF understands.  
	 * 
	 * @param authLevel string representation of the auth level
	 * @return PPF auth level from the string representation, defaults to PPF.AUTH_NONE if a 
	 * value that is not recognised if given
	 */
	protected int authLevelFromString(String authLevel) {		
		String level = authLevel.toUpperCase().trim();

		if(level.equals("NONE")) {
			return PPF.AUTH_NONE;
		} else if(level.equals("ANY")) {
			return PPF.AUTH_ANY;
		} else if(level.equals("ADMIN")){
			return PPF.AUTH_ADMIN;
		} else if(level.equals("MASTER")) {
			return PPF.AUTH_MASTER;
		} else if(level.equals("TRUSTED")) {
			return PPF.AUTH_TRUSTED;
		}
		
		return PPF.AUTH_NONE;
	}

	/**
	 * Send info that login has AUTHed with the system and the channel they are in.  This method is 
	 * called repeatedly from PPF when someone is authed with each channel they are in.
	 *    
	 * Information, such as, what auth level can be found from the isAuth(login, authLevel) method in PPF.
	 * 
	 * @param login the login that just AUTHed
	 * @param channel the channel that 
	 */
	protected void onAuth(String login, String channel) {}
	
	/////////////////////////////////////////////////////////////////
	//
	// PircBot onXXX methods
	//
	/////////////////////////////////////////////////////////////////
    
    /**
     * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onAction(java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)">PircBot onAction</a>
     */
    protected void onAction(String sender, String login, String hostname, 
        String target, String action) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onChannelInfo(java.lang.String,%20int,%20java.lang.String)">PircBot onChannelInfo</a>
	 */
    protected void onChannelInfo(String channel, int userCount, String topic) {}

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onConnect()">PircBot onConnect</a>
	 */
    protected void onConnect() {} 

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onDeop(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onDeop</a>
	 */
    protected void onDeop(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {} 

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onDeVoice(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onDeVoice</a>
	 */
    protected void onDeVoice(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onDisconnect()">PircBot onDisconnect</a>
	 */
    protected void onDisconnect() {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onFileTransferFinished(org.jibble.pircbot.DccFileTransfer,%20java.lang.Exception)">PircBot onFileTransferFinished</a>
	 */
    protected void onFileTransferFinished(DccFileTransfer transfer, Exception e) {} 

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onFinger(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onFinger</a>
	 */
    protected void onFinger(String sourceNick, String sourceLogin, 
        String sourceHostname, String target) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onIncomingChatRequest(org.jibble.pircbot.DccChat)">PircBot onIncomingChatRequest</a>
	 */
    protected void onIncomingChatRequest(DccChat chat) {}                 

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onIncomingFileTransfer(org.jibble.pircbot.DccFileTransfer)">PircBot onIncomingFileTransfer</a>
	 */
    protected void onIncomingFileTransfer(DccFileTransfer transfer) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onInvite(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onInvite</a>
	 */
    protected void onInvite(String targetNick, String sourceNick, String sourceLogin, 
        String sourceHostname, String channel) {}
                
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onJoin(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onJoin</a>
	 */
    protected void onJoin(String channel, String sender, String login, 
        String hostname) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onKick(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onKick</a>
	 */
    protected void onKick(String channel, String kickerNick, String kickerLogin, 
        String kickerHostname, String recipientNick, String reason) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onMessage(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onMessage</a>
	 */
    protected void onMessage(String channel, String sender, String login,
        String hostname, String message) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onMode(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onMode</a>
	 */
    protected void onMode(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String mode) {} 

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onNickChange(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onNickChange</a>
	 */
    protected void onNickChange(String oldNick, String login, String hostname, 
        String newNick) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onNotice(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onNotice</a>
	 */
    protected void onNotice(String sourceNick, String sourceLogin, String sourceHostname, 
        String target, String notice) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onOp(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onOp</a>
	 */
    protected void onOp(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onPart(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onPart</a>
	 */
    protected void onPart(String channel, String sender, String login, String hostname) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onPing(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onPing</a>
	 */
    protected void onPing(String sourceNick, String sourceLogin, 
        String sourceHostname, String target, String pingValue) {}

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onPrivateMessage(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onPrivateMessage</a>
	 */
    protected void onPrivateMessage(String sender, String login, String hostname, 
        String message) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onQuit(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onQuit</a>
	 */
    protected void onQuit(String sourceNick, String sourceLogin, String sourceHostname, 
        String reason) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveChannelBan(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveChannelBan</a>
	 */
    protected void onRemoveChannelBan(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String hostmask) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveChannelKey(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveChannelKey</a>
	 */
    protected void onRemoveChannelKey(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String key) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveChannelLimit(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveChannelLimit</a>
	 */
    protected void onRemoveChannelLimit(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveInviteOnly(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveInviteOnly</a>
	 */
    protected void onRemoveInviteOnly(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveModerated(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveModerated</a>
	 */
    protected void onRemoveModerated(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveNoExternalMessages(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveNoExternalMessages</a>
	 */
    protected void onRemoveNoExternalMessages(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemovePrivate(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemovePrivate</a>
	 */
    protected void onRemovePrivate(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {} 
 
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveSecret(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveSecret</a>
	 */
    protected void onRemoveSecret(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onRemoveTopicProtection(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onRemoveTopicProtection</a>
	 */
    protected void onRemoveTopicProtection(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onServerPing(java.lang.String)">PircBot onServerPing</a>
	 */
    protected void onServerPing(String response) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onServerResponse(int,%20java.lang.String)">PircBot onServerResponse</a>
	 */
    protected void onServerResponse(int code, String response) {}

	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetChannelBan(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetChannelBan</a>
	 */
    protected void onSetChannelBan(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String hostmask) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetChannelKey(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetChannelKey</a>
	 */
    protected void onSetChannelKey(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, String key) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetChannelLimit(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20int)">PircBot onSetChannelLimit</a>
	 */
    protected void onSetChannelLimit(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname, int limit) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetInviteOnly(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetInviteOnly</a>
	 */
    protected void onSetInviteOnly(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetModerated(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetModerated</a>
	 */
    protected void onSetModerated(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetNoExternalMessages(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetNoExternalMessages</a>
	 */
    protected void onSetNoExternalMessages(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetPrivate(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetPrivate</a>
	 */
    protected void onSetPrivate(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetSecret(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetSecret</a>
	 */
    protected void onSetSecret(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onSetTopicProtection(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onSetTopicProtection</a>
	 */
    protected void onSetTopicProtection(String channel, String sourceNick, 
        String sourceLogin, String sourceHostname) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onTime(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onTime</a>
	 */
    protected void onTime(String sourceNick, String sourceLogin, 
        String sourceHostname, String target) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onTopic(java.lang.String,%20java.lang.String,%20java.lang.String,%20long,%20boolean)">PircBot onTopic</a>
	 */
    protected void onTopic(String channel, String topic, String setBy, long date, 
        boolean changed) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onUnknown(java.lang.String)">PircBot onUnknown</a>
	 */
    protected void onUnknown(String line) {}
    
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onUserList(java.lang.String,%20org.jibble.pircbot.User[])">PircBot onUserList</a>
	 */
    protected void onUserList(String channel, User[] users) {} 
  
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onUserMode(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onUserMode</a>
	 */
    protected void onUserMode(String targetNick, String sourceNick, 
        String sourceLogin, String sourceHostname, String mode) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onVersion(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onVersion</a>
	 */
    protected void onVersion(String sourceNick, String sourceLogin, 
        String sourceHostname, String target) {}
        
	/**
	 * JavaDocs for this method are from: <a href="http://www.jibble.org/javadocs/pircbot/org/jibble/pircbot/PircBot.html#onVoice(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String)">PircBot onVoice</a>
	 */
    protected void onVoice(String channel, String sourceNick, String sourceLogin, 
        String sourceHostname, String recipient) {}
        
}
